"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations
from typing import Dict, List, Optional, TYPE_CHECKING, Any, Tuple, Union

from .utils import _get_as_snowflake

if TYPE_CHECKING:
    from aiohttp import ClientResponse, ClientWebSocketResponse
    from requests import Response

    _ResponseType = Union[ClientResponse, Response]

__all__ = (
    'DiscordException',
    'ClientException',
    'GatewayNotFound',
    'HTTPException',
    'RateLimited',
    'Forbidden',
    'NotFound',
    'DiscordServerError',
    'InvalidData',
    'AuthFailure',
    'LoginFailure',
    'ConnectionClosed',
    'CaptchaRequired',
)


class DiscordException(Exception):
    """Base exception class for discord.py

    Ideally speaking, this could be caught to handle any exceptions raised from this library.
    """

    pass


class ClientException(DiscordException):
    """Exception that's raised when an operation in the :class:`Client` fails.

    These are usually for exceptions that happened due to user input.
    """

    pass


class GatewayNotFound(DiscordException):
    """An exception that is raised when the gateway for Discord could not be found"""

    def __init__(self):
        message = 'The gateway to connect to Discord was not found.'
        super().__init__(message)


def _flatten_error_dict(d: Dict[str, Any], key: str = '') -> Dict[str, str]:
    items: List[Tuple[str, str]] = []

    if '_errors' in d:
        items.append(('miscallenous', ' '.join(x.get('message', '') for x in d['_errors'])))
        d.pop('_errors')

    for k, v in d.items():
        new_key = key + '.' + k if key else k

        if isinstance(v, dict):
            try:
                _errors: List[Dict[str, Any]] = v['_errors']
            except KeyError:
                items.extend(_flatten_error_dict(v, new_key).items())
            else:
                items.append((new_key, ' '.join(x.get('message', '') for x in _errors)))
        else:
            items.append((new_key, v))

    return dict(items)


class HTTPException(DiscordException):
    """Exception that's raised when an HTTP request operation fails.

    Attributes
    ------------
    response: :class:`aiohttp.ClientResponse`
        The response of the failed HTTP request. This is an
        instance of :class:`aiohttp.ClientResponse`. In some cases
        this could also be a :class:`requests.Response`.
    text: :class:`str`
        The text of the error. Could be an empty string.
    status: :class:`int`
        The status code of the HTTP request.
    code: :class:`int`
        The Discord specific error code for the failure.
    json: :class:`dict`
        The raw error JSON.

        .. versionadded:: 2.0
    payment_id: Optional[:class:`int`]
        The ID of the payment that requires verification to continue.

        .. versionadded:: 2.0
    """

    def __init__(self, response: _ResponseType, message: Optional[Union[str, Dict[str, Any]]]):
        self.response: _ResponseType = response
        self.status: int = response.status  # type: ignore # This attribute is filled by the library even if using requests
        self.code: int
        self.text: str
        self.json: Dict[str, Any]
        self.payment_id: Optional[int]
        if isinstance(message, dict):
            self.json = message
            self.code = message.get('code', 0)
            base = message.get('message', '')
            errors = message.get('errors')
            if errors:
                errors = _flatten_error_dict(errors)
                helpful = '\n'.join('In %s: %s' % t for t in errors.items())
                self.text = base + '\n' + helpful
            else:
                self.text = base
            self.payment_id = _get_as_snowflake(message, 'payment_id')
        else:
            self.text = message or ''
            self.code = 0
            self.json = {}
            self.payment_id = None

        fmt = '{0.status} {0.reason} (error code: {1})'
        if len(self.text):
            fmt += ': {2}'

        super().__init__(fmt.format(self.response, self.code, self.text))


class RateLimited(DiscordException):
    """Exception that's raised for when status code 429 occurs
    and the timeout is greater than the configured maximum using
    the ``max_ratelimit_timeout`` parameter in :class:`Client`.

    This is not raised during global ratelimits.

    Since sometimes requests are halted pre-emptively before they're
    even made, this **does not** subclass :exc:`HTTPException`.

    .. versionadded:: 2.0

    Attributes
    ------------
    retry_after: :class:`float`
        The amount of seconds that the client should wait before retrying
        the request.
    """

    def __init__(self, retry_after: float):
        self.retry_after = retry_after
        super().__init__(f'Too many requests. Retry in {retry_after:.2f} seconds.')


class Forbidden(HTTPException):
    """Exception that's raised for when status code 403 occurs.

    Subclass of :exc:`HTTPException`
    """

    pass


class NotFound(HTTPException):
    """Exception that's raised for when status code 404 occurs.

    Subclass of :exc:`HTTPException`
    """

    pass


class DiscordServerError(HTTPException):
    """Exception that's raised for when a 500 range status code occurs.

    Subclass of :exc:`HTTPException`.

    .. versionadded:: 1.5
    """

    pass


class CaptchaRequired(HTTPException):
    """Exception that's raised when a captcha is required and isn't handled.

    Subclass of :exc:`HTTPException`.

    .. versionadded:: 2.0
    """

    def __init__(self, response: _ResponseType, message: Dict[str, Any]):
        super().__init__(response, {'code': -1, 'message': 'Captcha required'})
        self.json = message


class InvalidData(ClientException):
    """Exception that's raised when the library encounters unknown
    or invalid data from Discord.
    """

    pass


class LoginFailure(ClientException):
    """Exception that's raised when the :meth:`Client.login` function
    fails to log you in from improper credentials or some other misc.
    failure.
    """

    pass


AuthFailure = LoginFailure


class ConnectionClosed(ClientException):
    """Exception that's raised when the gateway connection is
    closed for reasons that could not be handled internally.

    Attributes
    -----------
    code: :class:`int`
        The close code of the websocket.
    reason: :class:`str`
        The reason provided for the closure.
    """

    def __init__(self, socket: ClientWebSocketResponse, *, code: Optional[int] = None):
        # This exception is just the same exception except
        # reconfigured to subclass ClientException for users
        self.code: int = code or socket.close_code or -1
        # aiohttp doesn't seem to consistently provide close reason
        self.reason: str = ''
        super().__init__(f'WebSocket closed with {self.code}')



"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import Dict, List, Literal, Optional, TypedDict, Union
from typing_extensions import NotRequired

from .application import PartialApplication, StoreAsset
from .entitlements import Entitlement
from .guild import PartialGuild
from .library import LibraryApplication
from .snowflake import Snowflake
from .user import PartialUser

LOCALIZED_STR = Union[str, Dict[str, str]]


class StoreNote(TypedDict):
    content: str
    user: Optional[PartialUser]


class SystemRequirement(TypedDict, total=False):
    ram: int
    disk: int
    operating_system_version: LOCALIZED_STR
    cpu: LOCALIZED_STR
    gpu: LOCALIZED_STR
    sound_card: LOCALIZED_STR
    directx: LOCALIZED_STR
    network: LOCALIZED_STR
    notes: LOCALIZED_STR


class SystemRequirements(TypedDict, total=False):
    minimum: SystemRequirement
    recommended: SystemRequirement


class CarouselItem(TypedDict, total=False):
    asset_id: Snowflake
    youtube_video_id: str


class StoreListing(TypedDict):
    id: Snowflake
    summary: NotRequired[LOCALIZED_STR]
    description: NotRequired[LOCALIZED_STR]
    tagline: NotRequired[LOCALIZED_STR]
    flavor_text: NotRequired[str]
    published: NotRequired[bool]
    entitlement_branch_id: NotRequired[Snowflake]
    staff_notes: NotRequired[StoreNote]
    guild: NotRequired[PartialGuild]
    assets: NotRequired[List[StoreAsset]]
    carousel_items: NotRequired[List[CarouselItem]]
    preview_video: NotRequired[StoreAsset]
    header_background: NotRequired[StoreAsset]
    hero_background: NotRequired[StoreAsset]
    hero_video: NotRequired[StoreAsset]
    box_art: NotRequired[StoreAsset]
    thumbnail: NotRequired[StoreAsset]
    header_logo_light_theme: NotRequired[StoreAsset]
    header_logo_dark_theme: NotRequired[StoreAsset]
    sku: SKU
    child_skus: NotRequired[List[SKU]]
    alternative_skus: NotRequired[List[SKU]]


class SKUPrice(TypedDict):
    currency: str
    amount: int
    sale_amount: NotRequired[Optional[int]]
    sale_percentage: NotRequired[int]
    premium: NotRequired[bool]


class ContentRating(TypedDict):
    rating: int
    descriptors: List[int]


class PartialSKU(TypedDict):
    id: Snowflake
    type: Literal[1, 2, 3, 4, 5, 6]
    premium: bool
    preorder_release_date: Optional[str]
    preorder_released_at: Optional[str]


class SKU(PartialSKU):
    id: Snowflake
    type: Literal[1, 2, 3, 4, 5, 6]
    name: LOCALIZED_STR
    summary: NotRequired[LOCALIZED_STR]
    legal_notice: NotRequired[LOCALIZED_STR]
    slug: str
    dependent_sku_id: Optional[Snowflake]
    application_id: Snowflake
    application: NotRequired[PartialApplication]
    flags: int
    price_tier: NotRequired[int]
    price: NotRequired[Union[SKUPrice, Dict[str, int]]]
    sale_price_tier: NotRequired[int]
    sale_price: NotRequired[Dict[str, int]]
    access_level: Literal[1, 2, 3]
    features: List[int]
    locales: NotRequired[List[str]]
    genres: NotRequired[List[int]]
    available_regions: NotRequired[List[str]]
    content_rating_agency: NotRequired[Literal[1, 2]]
    content_rating: NotRequired[ContentRating]
    content_ratings: NotRequired[Dict[Literal[1, 2], ContentRating]]
    system_requirements: NotRequired[Dict[Literal[1, 2, 3], SystemRequirements]]
    release_date: Optional[str]
    preorder_release_date: NotRequired[Optional[str]]
    preorder_released_at: NotRequired[Optional[str]]
    external_purchase_url: NotRequired[str]
    premium: NotRequired[bool]
    restricted: NotRequired[bool]
    exclusive: NotRequired[bool]
    show_age_gate: bool
    bundled_skus: NotRequired[List[SKU]]
    manifest_labels: Optional[List[Snowflake]]


class SKUPurchase(TypedDict):
    entitlements: List[Entitlement]
    library_applications: NotRequired[List[LibraryApplication]]
    gift_code: NotRequired[str]



from .utils.objects import MessageContent, AuthorProfile
from .utils.payloads import StartSocket
from threading import Thread
from .utils import payloads
import websocket
import logging
import inspect
import json
import time
import os



_log = logging.getLogger(__name__)


class SocketDiscord:
    def __init__(self, client):
        self.events = {}
        self.ws = None
        self.token = client.token
        self.client = client
        
        self.session_id = None
        

    def on(self, event):
        if payloads.FormatEvent(event):
          event = payloads.FormatEvent(event)
        def event_ready(handler):
            if event not in self.events:
                self.events[event] = []
            self.events[event].append(handler)
            return handler
        return event_ready

    def trig(self, event, *args, **kwargs):
        if event in self.events:
            for handler in self.events[event]:
                try:
                  if len(inspect.signature(handler).parameters) == 0:
                    handler()
                  else:
                     handler(*args, **kwargs)
                except Exception as g:
                   _log.critical(str(g))

    def send_json(self, payload):
        self.ws.send(json.dumps(payload))


    def receive_json(self):
      response = self.ws.recv()
      if response:
        return json.loads(response)
        
    def send_and_receive_json(self, payload):
      self.ws.send(json.dumps(payload))
      res = json.loads(self.ws.recv())
      return res

    def heartbeat(self):
      timer = 40
      payload = {
      "op": 1,
      "d": None
    }
      while True:
        time.sleep(timer)
        self.send_json(payload)

    def start_ws(self):
      self.ws = websocket.WebSocket()
      self.ws.connect('wss://gateway.discord.gg/?encoding=json')
      self.send_json(StartSocket(self.token))
      Thread(target=self.event_manager).start()

    def get_object(self, data):
      if data['t'] == "READY":
        self.session_id = data.get('d').get('session_id')
        if not self.client.no_warns:
          print(f"{self.client.var_} - Session ID: {self.session_id}")
        
      if data['t'] == 'MESSAGE_CREATE':
        if data['d']['content'].split()[0] in self.client.commands:
          try:
              arg = data['d']['content'].split()[1:]
              require_params = len(inspect.signature(self.client.commands[data['d']['content'].split()[0]]).parameters)-1
          except Exception as g:
            print(g)
          if len(arg) < require_params:
            while len(arg) < require_params:
              arg.append(None)
              
          return self.client.emit_command(data['d']['content'].split()[0], MessageContent(data, self.client).MessageContent, *arg[:require_params])
        return MessageContent(data, self.client).MessageContent

      if data['t'] == 'GUILD_MEMBER_LIST_UPDATE':
        if data['s'] == 3:
          try:
            for item in data['d']['ops'][0]['items']:
              if item.get('member'):
                self.client.online_members.append(AuthorProfile(item['member']['user']).AuthorProfile)
          except:self.client.online_members = []
      
      return data
      
    def event_manager(self):
      while True:
        event = self.receive_json()
        try:
            self.trig('all', event)
            self.trig(event['t'], self.get_object(event))

        except:
            self.trig('others', event)


"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Optional

from .mixins import Hashable
from .subscriptions import Subscription
from .utils import parse_time, utcnow

if TYPE_CHECKING:
    from .abc import Snowflake
    from .guild import Guild
    from .state import ConnectionState
    from .types.subscriptions import (
        PremiumGuildSubscription as PremiumGuildSubscriptionPayload,
        PremiumGuildSubscriptionSlot as PremiumGuildSubscriptionSlotPayload,
        PremiumGuildSubscriptionCooldown as PremiumGuildSubscriptionCooldownPayload,
    )

__all__ = (
    'PremiumGuildSubscription',
    'PremiumGuildSubscriptionSlot',
    'PremiumGuildSubscriptionCooldown',
)


class PremiumGuildSubscription(Hashable):
    """Represents a premium guild subscription (boost).

    .. container:: operations

        .. describe:: x == y

            Checks if two premium guild subscriptions are equal.

        .. describe:: x != y

            Checks if two premium guild subscriptions are not equal.

        .. describe:: hash(x)

            Returns the premium guild subscription's hash.

    .. versionadded:: 2.0

    Attributes
    ------------
    id: :class:`int`
        The ID of the guild premium subscription.
    guild_id: :class:`int`
        The ID of the guild this guild premium subscription belongs to.
    user_id: :class:`int`
        The ID of the user this guild premium subscription belongs to.
    user: :class:`User`
        The user this guild premium subscription belongs to.
    ended: :class:`bool`
        Whether the guild premium subscription has ended.
    ends_at: Optional[:class:`datetime.datetime`]
        When the guild premium subscription ends.
    """

    def __init__(self, *, state: ConnectionState, data: PremiumGuildSubscriptionPayload):
        self._state = state
        self._update(data)

    def _update(self, data: PremiumGuildSubscriptionPayload):
        state = self._state

        self.id = int(data['id'])
        self.guild_id = int(data['guild_id'])
        self.user_id = int(data['user_id'])
        self.user = state.store_user(data['user']) if 'user' in data else state.user
        self.ended = data.get('ended', False)
        self.ends_at: Optional[datetime] = parse_time(data.get('ends_at'))

    def __repr__(self) -> str:
        return f'<PremiumGuildSubscription id={self.id} guild_id={self.guild_id} user_id={self.user_id} ended={self.ended}>'

    @property
    def guild(self) -> Optional[Guild]:
        """Optional[:class:`Guild`]: The guild this guild premium subscription belongs to, if available."""
        return self._state._get_guild(self.guild_id)

    @property
    def remaining(self) -> Optional[timedelta]:
        """Optional[:class:`datetime.timedelta`]: The remaining time for this guild premium subscription.

        This is ``None`` if the subscription is not ending.
        """
        if self.ends_at is None or self.ends_at <= utcnow():
            return None

        return self.ends_at - utcnow()

    async def delete(self) -> None:
        """|coro|

        Deletes this guild premium subscription.

        Raises
        -------
        Forbidden
            You do not have permissions to delete this guild premium subscription.
        HTTPException
            Deleting the guild premium subscription failed.
        """
        await self._state.http.delete_guild_subscription(self.guild_id, self.id)


class PremiumGuildSubscriptionSlot(Hashable):
    """Represents a premium guild subscription (boost) slot.

    This is a slot that can be used on a guild (to boost it).

    .. container:: operations

        .. describe:: x == y

            Checks if two subscription slots are equal.

        .. describe:: x != y

            Checks if two subscription slots are not equal.

        .. describe:: hash(x)

            Returns the subscription slot's hash.

    .. versionadded:: 2.0

    Attributes
    ------------
    id: :class:`int`
        The ID of the guild subscription slot.
    subscription_id: :class:`int`
        The ID of the guild subscription this slot belongs to.
    canceled: :class:`bool`
        Whether the slot is canceled.
    cooldown_ends_at: Optional[:class:`datetime.datetime`]
        When the cooldown for this guild subscription slot ends.
    premium_guild_subscription: Optional[:class:`PremiumGuildSubscription`]
        The subscription this slot belongs to.
    """

    __slots__ = (
        'id',
        'subscription_id',
        'canceled',
        'cooldown_ends_at',
        'premium_guild_subscription',
        '_state',
    )

    def __init__(self, *, state: ConnectionState, data: PremiumGuildSubscriptionSlotPayload):
        self._state = state
        self._update(data)

    def _update(self, data: PremiumGuildSubscriptionSlotPayload):
        self.id = int(data['id'])
        self.subscription_id = int(data['subscription_id'])
        self.canceled = data.get('canceled', False)
        self.cooldown_ends_at: Optional[datetime] = parse_time(data.get('cooldown_ends_at'))

        premium_guild_subscription = data.get('premium_guild_subscription')
        self.premium_guild_subscription: Optional[PremiumGuildSubscription] = (
            PremiumGuildSubscription(state=self._state, data=premium_guild_subscription)
            if premium_guild_subscription is not None
            else None
        )

    def __repr__(self) -> str:
        return f'<PremiumGuildSubscriptionSlot id={self.id} subscription_id={self.subscription_id} canceled={self.canceled}>'

    def is_available(self) -> bool:
        """:class:`bool`: Indicates if the slot is available for use."""
        return not self.premium_guild_subscription and not self.is_on_cooldown()

    def is_on_cooldown(self) -> bool:
        """:class:`bool`: Indicates if the slot is on cooldown."""
        return self.cooldown_ends_at is not None and self.cooldown_ends_at > utcnow()

    @property
    def cancelled(self) -> bool:
        """:class:`bool`: Whether the slot is cancelled.

        This is an alias of :attr:`canceled`.
        """
        return self.canceled

    @property
    def cooldown_remaining(self) -> Optional[timedelta]:
        """Optional[:class:`datetime.timedelta`]: The cooldown remaining for this boost slot.

        This is ``None`` if the cooldown has ended.
        """
        if self.cooldown_ends_at is None or self.cooldown_ends_at <= utcnow():
            return None

        return self.cooldown_ends_at - utcnow()

    async def subscription(self) -> Subscription:
        """|coro|

        Retrieves the subscription this guild subscription slot is attached to.

        Raises
        ------
        NotFound
            You cannot access this subscription.
        HTTPException
            Fetching the subscription failed.

        Returns
        -------
        :class:`Subscription`
            The retrieved subscription, if applicable.
        """
        data = await self._state.http.get_subscription(self.subscription_id)
        return Subscription(data=data, state=self._state)

    async def apply(self, guild: Snowflake) -> PremiumGuildSubscription:
        """|coro|

        Applies the premium guild subscription slot to a guild.

        Parameters
        -----------
        guild: :class:`Guild`
            The guild to apply the slot to.

        Raises
        -------
        HTTPException
            Applying the slot failed.

        Returns
        --------
        :class:`PremiumGuildSubscription`
            The premium guild subscription that was created.
        """
        state = self._state
        data = await state.http.apply_guild_subscription_slots(guild.id, (self.id,))
        return PremiumGuildSubscription(state=state, data=data[0])

    async def cancel(self) -> None:
        """|coro|

        Cancels the guild subscription slot.

        Raises
        -------
        HTTPException
            Cancelling the slot failed.
        """
        data = await self._state.http.cancel_guild_subscription_slot(self.id)
        self._update(data)

    async def uncancel(self) -> None:
        """|coro|

        Uncancels the guild subscription slot.

        Raises
        -------
        HTTPException
            Uncancelling the slot failed.
        """
        data = await self._state.http.uncancel_guild_subscription_slot(self.id)
        self._update(data)


class PremiumGuildSubscriptionCooldown:
    """Represents a premium guild subscription cooldown.

    This is a cooldown that is applied to your guild subscription slot changes (boosting and unboosting).

    .. versionadded:: 2.0

    Attributes
    ------------
    ends_at: :class:`datetime.datetime`
        When the cooldown resets.
    limit: :class:`int`
        The maximum number of changes that can be made before the cooldown is applied.
    remaining: :class:`int`
        The number of changes remaining before the cooldown is applied.
    """

    def __init__(self, *, state: ConnectionState, data: PremiumGuildSubscriptionCooldownPayload):
        self._state = state
        self._update(data)

    def _update(self, data: PremiumGuildSubscriptionCooldownPayload):
        self.ends_at: datetime = parse_time(data['ends_at'])
        self.limit = data['limit']
        self.remaining = data.get('remaining', 0)



"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, List, Optional, Union

from .enums import PaymentSourceType, try_enum
from .flags import PromotionFlags
from .mixins import Hashable
from .subscriptions import SubscriptionTrial
from .utils import _get_as_snowflake, parse_time, utcnow

if TYPE_CHECKING:
    from .state import ConnectionState
    from .types.promotions import (
        ClaimedPromotion as ClaimedPromotionPayload,
        Promotion as PromotionPayload,
        TrialOffer as TrialOfferPayload,
        PricingPromotion as PricingPromotionPayload,
    )

__all__ = (
    'Promotion',
    'TrialOffer',
    'PricingPromotion',
)


class Promotion(Hashable):
    """Represents a Discord promotion.

    .. container:: operations

        .. describe:: x == y

            Checks if two promotions are equal.

        .. describe:: x != y

            Checks if two promotions are not equal.

        .. describe:: hash(x)

            Returns the promotion's hash.

        .. describe:: str(x)

            Returns the outbound promotion's name.

    .. versionadded:: 2.0

    Attributes
    ----------
    id: :class:`int`
        The promotion ID.
    trial_id: Optional[:class:`int`]
        The trial ID of the inbound promotion, if applicable.
    starts_at: :class:`datetime.datetime`
        When the promotion starts.
    ends_at: :class:`datetime.datetime`
        When the promotion ends.
    claimed_at: Optional[:class:`datetime.datetime`]
        When the promotion was claimed.
        Only available for claimed promotions.
    code: Optional[:class:`str`]
        The promotion's claim code. Only available for claimed promotions.
    outbound_title: :class:`str`
        The title of the outbound promotion.
    outbound_description: :class:`str`
        The description of the outbound promotion.
    outbound_link: :class:`str`
        The redemption page of the outbound promotion, used to claim it.
    outbound_restricted_countries: List[:class:`str`]
        The countries that the outbound promotion is not available in.
    inbound_title: Optional[:class:`str`]
        The title of the inbound promotion. This is usually Discord Nitro.
    inbound_description: Optional[:class:`str`]
        The description of the inbound promotion.
    inbound_link: Optional[:class:`str`]
        The Discord help center link of the inbound promotion.
    inbound_restricted_countries: List[:class:`str`]
        The countries that the inbound promotion is not available in.
    terms_and_conditions: :class:`str`
        The terms and conditions of the promotion.
    """

    __slots__ = (
        'id',
        'trial_id',
        'starts_at',
        'ends_at',
        'claimed_at',
        'code',
        'outbound_title',
        'outbound_description',
        'outbound_link',
        'outbound_restricted_countries',
        'inbound_title',
        'inbound_description',
        'inbound_link',
        'inbound_restricted_countries',
        'terms_and_conditions',
        '_flags',
        '_state',
    )

    def __init__(self, *, data: Union[PromotionPayload, ClaimedPromotionPayload], state: ConnectionState) -> None:
        self._state = state
        self._update(data)

    def __str__(self) -> str:
        return self.outbound_title

    def __repr__(self) -> str:
        return f'<Promotion id={self.id} title={self.outbound_title!r}>'

    def _update(self, data: Union[PromotionPayload, ClaimedPromotionPayload]) -> None:
        promotion: PromotionPayload = data.get('promotion', data)  # type: ignore

        self.id: int = int(promotion['id'])
        self.trial_id: Optional[int] = _get_as_snowflake(promotion, 'trial_id')
        self.starts_at: datetime = parse_time(promotion['start_date'])
        self.ends_at: datetime = parse_time(promotion['end_date'])
        self.claimed_at: Optional[datetime] = parse_time(data.get('claimed_at'))
        self.code: Optional[str] = data.get('code')
        self._flags: int = promotion.get('flags', 0)

        self.outbound_title: str = promotion['outbound_title']
        self.outbound_description: str = promotion['outbound_redemption_modal_body']
        self.outbound_link: str = promotion.get(
            'outbound_redemption_page_link',
            promotion.get('outbound_redemption_url_format', '').replace('{code}', self.code or '{code}'),
        )
        self.outbound_restricted_countries: List[str] = promotion.get('outbound_restricted_countries', [])
        self.inbound_title: Optional[str] = promotion.get('inbound_header_text')
        self.inbound_description: Optional[str] = promotion.get('inbound_body_text')
        self.inbound_link: Optional[str] = promotion.get('inbound_help_center_link')
        self.inbound_restricted_countries: List[str] = promotion.get('inbound_restricted_countries', [])
        self.terms_and_conditions: str = promotion['outbound_terms_and_conditions']

    @property
    def flags(self) -> PromotionFlags:
        """:class:`PromotionFlags`: Returns the promotion's flags."""
        return PromotionFlags._from_value(self._flags)

    def is_claimed(self) -> bool:
        """:class:`bool`: Checks if the promotion has been claimed.

        Only accurate if the promotion was fetched from :meth:`Client.promotions` with ``claimed`` set to ``True`` or :meth:`claim` was just called.
        """
        return self.claimed_at is not None

    def is_active(self) -> bool:
        """:class:`bool`: Checks if the promotion is active."""
        return self.starts_at <= utcnow() <= self.ends_at

    async def claim(self) -> str:
        """|coro|

        Claims the promotion.

        Sets :attr:`claimed_at` and :attr:`code`.

        Raises
        ------
        Forbidden
            You are not allowed to claim the promotion.
        HTTPException
            Claiming the promotion failed.

        Returns
        -------
        :class:`str`
            The claim code for the outbound promotion.
        """
        data = await self._state.http.claim_promotion(self.id)
        self._update(data)
        return data['code']


class TrialOffer(Hashable):
    """Represents a Discord user trial offer.

    .. container:: operations

        .. describe:: x == y

            Checks if two trial offers are equal.

        .. describe:: x != y

            Checks if two trial offers are not equal.

        .. describe:: hash(x)

            Returns the trial offer's hash.

    .. versionadded:: 2.0

    Attributes
    ----------
    id: :class:`int`
        The ID of the trial offer.
    expires_at: :class:`datetime.datetime`
        When the trial offer expires.
    trial_id: :class:`int`
        The ID of the trial.
    trial: :class:`SubscriptionTrial`
        The trial offered.
    """

    __slots__ = (
        'id',
        'expires_at',
        'trial_id',
        'trial',
        '_state',
    )

    def __init__(self, *, data: TrialOfferPayload, state: ConnectionState) -> None:
        self._state = state

        self.id: int = int(data['id'])
        self.expires_at: datetime = parse_time(data['expires_at'])
        self.trial_id: int = int(data['trial_id'])
        self.trial: SubscriptionTrial = SubscriptionTrial(data['subscription_trial'])

    def __repr__(self) -> str:
        return f'<TrialOffer id={self.id} trial={self.trial!r}>'

    async def ack(self) -> None:
        """|coro|

        Acknowledges the trial offer.

        Raises
        ------
        HTTPException
            Acknowledging the trial offer failed.
        """
        await self._state.http.ack_trial_offer(self.id)


class PricingPromotion:
    """Represents a Discord localized pricing promotion.

    .. versionadded:: 2.0

    Attributes
    ----------
    subscription_plan_id: :class:`int`
        The ID of the subscription plan the promotion is for.
    country_code: :class:`str`
        The country code the promotion applies to.
    payment_source_types: List[:class:`PaymentSourceType`]
        The payment source types the promotion is restricted to.
    amount: :class:`int`
        The discounted price of the subscription plan.
    currency: :class:`str`
        The currency of the discounted price.
    """

    __slots__ = (
        'subscription_plan_id',
        'country_code',
        'payment_source_types',
        'amount',
        'currency',
    )

    def __init__(self, *, data: PricingPromotionPayload) -> None:
        self.subscription_plan_id: int = int(data['plan_id'])
        self.country_code: str = data['country_code']
        self.payment_source_types: List[PaymentSourceType] = [
            try_enum(PaymentSourceType, t) for t in data['payment_source_types']
        ]

        price = data['price']
        self.amount: int = price['amount']
        self.currency: str = price['currency']

    def __repr__(self) -> str:
        return f'<PricingPromotion plan_id={self.subscription_plan_id} country_code={self.country_code!r} amount={self.amount} currency={self.currency!r}>'



import yt_dlp as youtube_dl
from .. import discord
import asyncio
from threading import Thread


youtube_dl.utils.bug_reports_message = lambda: ''


ytdl_format_options = {
    'format': 'bestaudio/best',
    'outtmpl': '%(extractor)s-%(id)s-%(title)s.%(ext)s',
    'restrictfilenames': True,
    'noplaylist': True,
    'nocheckcertificate': True,
    'ignoreerrors': False,
    'logtostderr': False,
    'quiet': True,
    'no_warnings': True,
    'default_search': 'auto',
    'source_address': '0.0.0.0',
}

ffmpeg_options = {
    'options': '-vn',
}

ytdl = youtube_dl.YoutubeDL(ytdl_format_options)


class YTDLSource(discord.PCMVolumeTransformer):
    def __init__(self, source, *, data, volume=0.5):
        super().__init__(source, volume)

        self.data = data

        self.title = data.get('title')
        self.url = data.get('url')

    @classmethod
    async def from_url(cls, url, *, loop=None, stream=False):
        loop = loop or asyncio.get_event_loop()
        data = await loop.run_in_executor(None, lambda: ytdl.extract_info(url, download=not stream))

        if 'entries' in data:
            data = data['entries'][0]

        print(url)

        filename = data['url'] if stream else ytdl.prepare_filename(data)
        return cls(discord.FFmpegPCMAudio(filename, **ffmpeg_options), data=data)


"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Optional, Union

from .enums import (
    PaymentGateway,
    PaymentSourceType,
    try_enum,
)
from .flags import PaymentSourceFlags
from .mixins import Hashable
from .utils import MISSING

if TYPE_CHECKING:
    from datetime import date
    from typing_extensions import Self

    from .state import ConnectionState
    from .types.billing import (
        BillingAddress as BillingAddressPayload,
        PartialPaymentSource as PartialPaymentSourcePayload,
        PaymentSource as PaymentSourcePayload,
        PremiumUsage as PremiumUsagePayload,
    )

__all__ = (
    'BillingAddress',
    'PaymentSource',
    'PremiumUsage',
)


class BillingAddress:
    """Represents a billing address.

    .. container:: operations

        .. describe:: x == y

            Checks if two billing addresses are equal.

        .. describe:: x != y

            Checks if two billing addresses are not equal.

        .. describe:: hash(x)

            Returns the address' hash.

    .. versionadded:: 2.0

    Attributes
    ----------
    name: :class:`str`
        The payment source's name.
    address: :class:`str`
        The location's address.
    postal_code: Optional[:class:`str`]
        The location's postal code.
    city: :class:`str`
        The location's city.
    state: Optional[:class:`str`]
        The location's state or province.
    country: :class:`str`
        The location's country.
    email: Optional[:class:`str`]
        The email address associated with the payment source, if any.
    """

    __slots__ = ('_state', 'name', 'address', 'postal_code', 'city', 'state', 'country', 'email')

    def __init__(
        self,
        *,
        name: str,
        address: str,
        city: str,
        country: str,
        state: Optional[str] = None,
        postal_code: Optional[str] = None,
        email: Optional[str] = None,
        _state: Optional[ConnectionState] = None,
    ) -> None:
        self._state = _state

        self.name = name
        self.address = address
        self.postal_code = postal_code
        self.city = city
        self.state = state
        self.country = country
        self.email = email

    def __repr__(self) -> str:
        return f'<BillingAddress name={self.name!r} address={self.address!r} city={self.city!r} country={self.country!r}>'

    def __eq__(self, other: object) -> bool:
        return isinstance(other, BillingAddress) and self.to_dict() == other.to_dict()

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, BillingAddress):
            return True
        return self.to_dict() != other.to_dict()

    def __hash__(self) -> int:
        return hash(self.to_dict())

    @classmethod
    def from_dict(cls, data: BillingAddressPayload, state: ConnectionState) -> Self:
        address = '\n'.join(filter(None, (data['line_1'], data.get('line_2'))))
        return cls(
            _state=state,
            name=data['name'],
            address=address,
            postal_code=data.get('postal_code'),
            city=data['city'],
            state=data.get('state'),
            country=data['country'],
            email=data.get('email'),
        )

    def to_dict(self) -> dict:
        line1, _, line2 = self.address.partition('\n')
        data = {
            'name': self.name,
            'line_1': line1,
            'line_2': line2 or '',
            'city': self.city,
            'country': self.country,
        }
        if self.postal_code:
            data['postal_code'] = self.postal_code
        if self.state:
            data['state'] = self.state
        if self.email:
            data['email'] = self.email
        return data

    async def validate(self) -> str:
        """|coro|

        Validates the billing address.

        Raises
        ------
        TypeError
            The billing address does not have state attached.
        HTTPException
            The billing address is invalid.

        Returns
        -------
        :class:`str`
            The billing address token.
        """
        if self._state is None:
            raise TypeError('BillingAddress does not have state available')

        data = await self._state.http.validate_billing_address(self.to_dict())
        return data['token']


class PaymentSource(Hashable):
    """Represents a payment source.

    .. container:: operations

        .. describe:: x == y

            Checks if two payment sources are equal.

        .. describe:: x != y

            Checks if two payment sources are not equal.

        .. describe:: hash(x)

            Returns the source's hash.

    .. versionadded:: 2.0

    Attributes
    ----------
    id: :class:`int`
        The ID of the payment source.
    brand: Optional[:class:`str`]
        The brand of the payment source. This is only available for cards.
    country: Optional[:class:`str`]
        The country of the payment source. Not available in all contexts.
    partial_card_number: Optional[:class:`str`]
        The last four digits of the payment source. This is only available for cards.
    billing_address: Optional[:class:`BillingAddress`]
        The billing address of the payment source. Not available in all contexts.
    type: :class:`PaymentSourceType`
        The type of the payment source.
    payment_gateway: :class:`PaymentGateway`
        The payment gateway of the payment source.
    default: :class:`bool`
        Whether the payment source is the default payment source.
    invalid: :class:`bool`
        Whether the payment source is invalid.
    expires_at: Optional[:class:`datetime.date`]
        When the payment source expires. This is only available for cards.
    email: Optional[:class:`str`]
        The email address associated with the payment source, if any.
        This is only available for PayPal.
    bank: Optional[:class:`str`]
        The bank associated with the payment source, if any.
        This is only available for certain payment sources.
    username: Optional[:class:`str`]
        The username associated with the payment source, if any.
        This is only available for Venmo.
    """

    __slots__ = (
        '_state',
        'id',
        'brand',
        'country',
        'partial_card_number',
        'billing_address',
        'type',
        'payment_gateway',
        'default',
        'invalid',
        'expires_at',
        'email',
        'bank',
        'username',
        '_flags',
    )

    def __init__(self, *, data: Union[PaymentSourcePayload, PartialPaymentSourcePayload], state: ConnectionState) -> None:
        self._state = state
        self._update(data)

    def __repr__(self) -> str:
        return f'<PaymentSource id={self.id} type={self.type!r} country={self.country!r}>'

    def _update(self, data: Union[PaymentSourcePayload, PartialPaymentSourcePayload]) -> None:
        self.id: int = int(data['id'])
        self.brand: Optional[str] = data.get('brand')
        self.country: Optional[str] = data.get('country')
        self.partial_card_number: Optional[str] = data.get('last_4')
        self.billing_address: Optional[BillingAddress] = (
            BillingAddress.from_dict(data['billing_address'], state=self._state) if 'billing_address' in data else None  # type: ignore # ???
        )

        self.type: PaymentSourceType = try_enum(PaymentSourceType, data['type'])
        self.payment_gateway: PaymentGateway = try_enum(PaymentGateway, data['payment_gateway'])
        self.default: bool = data.get('default', False)
        self.invalid: bool = data['invalid']
        self._flags: int = data.get('flags', 0)

        month = data.get('expires_month')
        year = data.get('expires_year')
        self.expires_at: Optional[date] = datetime(year=year, month=month or 1, day=1).date() if year else None

        self.email: Optional[str] = data.get('email')
        self.bank: Optional[str] = data.get('bank')
        self.username: Optional[str] = data.get('username')

        if not self.country and self.billing_address:
            self.country = self.billing_address.country

        if not self.email and self.billing_address:
            self.email = self.billing_address.email

    @property
    def flags(self) -> PaymentSourceFlags:
        """:class:`PaymentSourceFlags`: Returns the payment source's flags."""
        return PaymentSourceFlags._from_value(self._flags)

    async def edit(
        self, *, billing_address: BillingAddress = MISSING, default: bool = MISSING, expires_at: date = MISSING
    ) -> None:
        """|coro|

        Edits the payment source.

        Parameters
        ----------
        billing_address: :class:`BillingAddress`
            The billing address of the payment source.
        default: :class:`bool`
            Whether the payment source is the default payment source.
        expires_at: :class:`datetime.date`
            When the payment source expires. This is only applicable to cards.

        Raises
        ------
        HTTPException
            Editing the payment source failed.
        """
        payload = {}
        if billing_address is not MISSING:
            payload['billing_address'] = billing_address.to_dict()
        if default is not MISSING:
            payload['default'] = default
        if expires_at is not MISSING:
            payload['expires_month'] = expires_at.month
            payload['expires_year'] = expires_at.year

        data = await self._state.http.edit_payment_source(self.id, payload)
        self._update(data)

    async def delete(self) -> None:
        """|coro|

        Deletes the payment source.

        Raises
        ------
        HTTPException
            Deleting the payment source failed.
        """
        await self._state.http.delete_payment_source(self.id)


class PremiumUsage:
    """Represents the usage of a user's premium perks.

    .. versionadded:: 2.0

    Attributes
    ----------
    sticker_sends: :class:`int`
        The number of premium sticker sends.
    animated_emojis: :class:`int`
        The number of animated emojis used.
    global_emojis: :class:`int`
        The number of global emojis used.
    large_uploads: :class:`int`
        The number of large uploads made.
    hd_streams: :class:`int`
        The number of HD streams.
    hd_hours_streamed: :class:`int`
        The number of hours streamed in HD.
    """

    __slots__ = (
        'sticker_sends',
        'animated_emojis',
        'global_emojis',
        'large_uploads',
        'hd_streams',
        'hd_hours_streamed',
    )

    def __init__(self, *, data: PremiumUsagePayload) -> None:
        self.sticker_sends: int = data['nitro_sticker_sends']['value']
        self.animated_emojis: int = data['total_animated_emojis']['value']
        self.global_emojis: int = data['total_global_emojis']['value']
        self.large_uploads: int = data['total_large_uploads']['value']
        self.hd_streams: int = data['total_hd_streams']['value']
        self.hd_hours_streamed: int = data['hd_hours_streamed']['value']



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union, overload

from .asset import Asset
from .colour import Colour
from .enums import ActivityType, ClientType, OperatingSystem, Status, try_enum
from .partial_emoji import PartialEmoji
from .utils import _get_as_snowflake, parse_time, parse_timestamp

__all__ = (
    'BaseActivity',
    'Activity',
    'Streaming',
    'Game',
    'Spotify',
    'CustomActivity',
    'Session',
)

"""If curious, this is the current schema for an activity.

It's fairly long so I will document it here:

All keys are optional.

state: str (max: 128),
details: str (max: 128)
timestamps: dict
    start: int (min: 1)
    end: int (min: 1)
assets: dict
    large_image: str (max: 32)
    large_text: str (max: 128)
    small_image: str (max: 32)
    small_text: str (max: 128)
party: dict
    id: str (max: 128),
    size: List[int] (max-length: 2)
        elem: int (min: 1)
secrets: dict
    match: str (max: 128)
    join: str (max: 128)
    spectate: str (max: 128)
instance: bool
application_id: str
name: str (max: 128)
url: str
type: int
sync_id: str
session_id: str
flags: int
buttons: list[str (max: 32)]

There are also activity flags which are mostly uninteresting for the library atm.

t.ActivityFlags = {
    INSTANCE: 1,
    JOIN: 2,
    SPECTATE: 4,
    JOIN_REQUEST: 8,
    SYNC: 16,
    PLAY: 32
}
"""

if TYPE_CHECKING:
    from typing_extensions import Self

    from .state import ConnectionState
    from .types.activity import (
        Activity as ActivityPayload,
        ActivityAssets,
        ActivityParty,
        ActivityTimestamps,
    )
    from .types.gateway import Session as SessionPayload


class BaseActivity:
    """The base activity that all user-settable activities inherit from.
    A user-settable activity is one that can be used in :meth:`Client.change_presence`.

    The following types currently count as user-settable:

    - :class:`Activity`
    - :class:`Game`
    - :class:`Streaming`
    - :class:`CustomActivity`

    Note that although these types are considered user-settable by the library,
    Discord typically ignores certain combinations of activity depending on
    what is currently set. This behaviour may change in the future so there are
    no guarantees on whether Discord will actually let you set these types.

    .. versionadded:: 1.3
    """

    __slots__ = ('_created_at',)

    def __init__(self, **kwargs: Any) -> None:
        self._created_at: Optional[float] = kwargs.pop('created_at', None)

    @property
    def created_at(self) -> Optional[datetime.datetime]:
        """Optional[:class:`datetime.datetime`]: When the user started doing this activity in UTC.

        .. versionadded:: 1.3
        """
        if self._created_at is not None:
            return datetime.datetime.fromtimestamp(self._created_at / 1000, tz=datetime.timezone.utc)

    def to_dict(self) -> ActivityPayload:
        raise NotImplementedError


class Activity(BaseActivity):
    """Represents an activity in Discord.

    This could be an activity such as streaming, playing, listening
    or watching.

    For memory optimisation purposes, some activities are offered in slimmed
    down versions:

    - :class:`Game`
    - :class:`Streaming`

    Attributes
    ------------
    application_id: Optional[:class:`int`]
        The application ID of the game.
    name: Optional[:class:`str`]
        The name of the activity.
    url: Optional[:class:`str`]
        A stream URL that the activity could be doing.
    type: :class:`ActivityType`
        The type of activity currently being done.
    state: Optional[:class:`str`]
        The user's current state. For example, "In Game".
    details: Optional[:class:`str`]
        The detail of the user's current activity.
    timestamps: :class:`dict`
        A dictionary of timestamps. It contains the following optional keys:

        - ``start``: Corresponds to when the user started doing the
          activity in milliseconds since Unix epoch.
        - ``end``: Corresponds to when the user will finish doing the
          activity in milliseconds since Unix epoch.

    assets: :class:`dict`
        A dictionary representing the images and their hover text of an activity.
        It contains the following optional keys:

        - ``large_image``: A string representing the ID for the large image asset.
        - ``large_text``: A string representing the text when hovering over the large image asset.
        - ``small_image``: A string representing the ID for the small image asset.
        - ``small_text``: A string representing the text when hovering over the small image asset.

    party: :class:`dict`
        A dictionary representing the activity party. It contains the following optional keys:

        - ``id``: A string representing the party ID.
        - ``size``: A list of up to two integer elements denoting (current_size, maximum_size).
    buttons: List[:class:`str`]
        A list of strings representing the labels of custom buttons shown in a rich presence.

        .. versionadded:: 2.0

    emoji: Optional[:class:`PartialEmoji`]
        The emoji that belongs to this activity.
    """

    __slots__ = (
        'state',
        'details',
        'timestamps',
        'assets',
        'party',
        'flags',
        'sync_id',
        'session_id',
        'type',
        'name',
        'url',
        'application_id',
        'emoji',
        'buttons',
    )

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.state: Optional[str] = kwargs.pop('state', None)
        self.details: Optional[str] = kwargs.pop('details', None)
        self.timestamps: ActivityTimestamps = kwargs.pop('timestamps', {})
        self.assets: ActivityAssets = kwargs.pop('assets', {})
        self.party: ActivityParty = kwargs.pop('party', {})
        self.application_id: Optional[int] = _get_as_snowflake(kwargs, 'application_id')
        self.name: Optional[str] = kwargs.pop('name', None)
        self.url: Optional[str] = kwargs.pop('url', None)
        self.flags: int = kwargs.pop('flags', 0)
        self.sync_id: Optional[str] = kwargs.pop('sync_id', None)
        self.session_id: Optional[str] = kwargs.pop('session_id', None)
        self.buttons: List[str] = kwargs.pop('buttons', [])

        activity_type = kwargs.pop('type', -1)
        self.type: ActivityType = (
            activity_type if isinstance(activity_type, ActivityType) else try_enum(ActivityType, activity_type)
        )

        emoji = kwargs.pop('emoji', None)
        self.emoji: Optional[PartialEmoji] = PartialEmoji.from_dict(emoji) if emoji is not None else None

    def __repr__(self) -> str:
        attrs = (
            ('type', self.type),
            ('name', self.name),
            ('url', self.url),
            ('details', self.details),
            ('application_id', self.application_id),
            ('session_id', self.session_id),
            ('emoji', self.emoji),
        )
        inner = ' '.join('%s=%r' % t for t in attrs)
        return f'<Activity {inner}>'

    def __eq__(self, other):
        return (
            isinstance(other, Activity)
            and other.type == self.type
            and other.name == self.name
            and other.url == self.url
            and other.emoji == self.emoji
            and other.state == self.state
            and other.session_id == self.session_id
            and other.sync_id == self.sync_id
            and other.start == self.start
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_dict(self) -> ActivityPayload:
        ret: Dict[str, Any] = {}
        for attr in self.__slots__:
            value = getattr(self, attr, None)
            if value is None:
                continue

            if isinstance(value, dict) and len(value) == 0:
                continue

            ret[attr] = value
        ret['type'] = int(self.type)
        if self.emoji:
            ret['emoji'] = self.emoji.to_dict()
        return ret  # type: ignore

    @property
    def start(self) -> Optional[datetime.datetime]:
        """Optional[:class:`datetime.datetime`]: When the user started doing this activity in UTC, if applicable."""
        try:
            timestamp = self.timestamps['start'] / 1000
        except KeyError:
            return None
        else:
            return datetime.datetime.fromtimestamp(timestamp, tz=datetime.timezone.utc)

    @property
    def end(self) -> Optional[datetime.datetime]:
        """Optional[:class:`datetime.datetime`]: When the user will stop doing this activity in UTC, if applicable."""
        try:
            timestamp = self.timestamps['end'] / 1000
        except KeyError:
            return None
        else:
            return datetime.datetime.fromtimestamp(timestamp, tz=datetime.timezone.utc)

    @property
    def large_image_url(self) -> Optional[str]:
        """Optional[:class:`str`]: Returns a URL pointing to the large image asset of this activity, if applicable."""
        try:
            large_image = self.assets['large_image']
        except KeyError:
            return None
        else:
            return self._image_url(large_image)

    @property
    def small_image_url(self) -> Optional[str]:
        """Optional[:class:`str`]: Returns a URL pointing to the small image asset of this activity, if applicable."""
        try:
            small_image = self.assets['small_image']
        except KeyError:
            return None
        else:
            return self._image_url(small_image)

    def _image_url(self, image: str) -> Optional[str]:
        if image.startswith('mp:'):
            return f'https://media.discordapp.net/{image[3:]}'
        elif self.application_id is not None:
            return Asset.BASE + f'/app-assets/{self.application_id}/{image}.png'

    @property
    def large_image_text(self) -> Optional[str]:
        """Optional[:class:`str`]: Returns the large image asset hover text of this activity, if applicable."""
        return self.assets.get('large_text', None)

    @property
    def small_image_text(self) -> Optional[str]:
        """Optional[:class:`str`]: Returns the small image asset hover text of this activity, if applicable."""
        return self.assets.get('small_text', None)


class Game(BaseActivity):
    """A slimmed down version of :class:`Activity` that represents a Discord game.

    This is typically displayed via **Playing** on the official Discord client.

    .. container:: operations

        .. describe:: x == y

            Checks if two games are equal.

        .. describe:: x != y

            Checks if two games are not equal.

        .. describe:: hash(x)

            Returns the game's hash.

        .. describe:: str(x)

            Returns the game's name.

    Parameters
    -----------
    name: :class:`str`
        The game's name.

    Attributes
    -----------
    name: :class:`str`
        The game's name.
    """

    __slots__ = ('name', '_end', '_start')

    def __init__(self, name: str, **extra: Any) -> None:
        super().__init__(**extra)
        self.name: str = name

        try:
            timestamps: ActivityTimestamps = extra['timestamps']
        except KeyError:
            self._start = 0
            self._end = 0
        else:
            self._start = timestamps.get('start', 0)
            self._end = timestamps.get('end', 0)

    @property
    def type(self) -> ActivityType:
        """:class:`ActivityType`: Returns the game's type. This is for compatibility with :class:`Activity`.

        It always returns :attr:`ActivityType.playing`.
        """
        return ActivityType.playing

    @property
    def start(self) -> Optional[datetime.datetime]:
        """Optional[:class:`datetime.datetime`]: When the user started playing this game in UTC, if applicable."""
        if self._start:
            return datetime.datetime.fromtimestamp(self._start / 1000, tz=datetime.timezone.utc)
        return None

    @property
    def end(self) -> Optional[datetime.datetime]:
        """Optional[:class:`datetime.datetime`]: When the user will stop playing this game in UTC, if applicable."""
        if self._end:
            return datetime.datetime.fromtimestamp(self._end / 1000, tz=datetime.timezone.utc)
        return None

    def __str__(self) -> str:
        return str(self.name)

    def __repr__(self) -> str:
        return f'<Game name={self.name!r}>'

    def to_dict(self) -> ActivityPayload:
        timestamps: Dict[str, Any] = {}
        if self._start:
            timestamps['start'] = self._start

        if self._end:
            timestamps['end'] = self._end

        return {
            'type': ActivityType.playing.value,
            'name': str(self.name),
            'timestamps': timestamps,  # type: ignore
        }

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Game) and other.name == self.name

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.name)


class Streaming(BaseActivity):
    """A slimmed down version of :class:`Activity` that represents a Discord streaming status.

    This is typically displayed via **Streaming** on the official Discord client.

    .. container:: operations

        .. describe:: x == y

            Checks if two streams are equal.

        .. describe:: x != y

            Checks if two streams are not equal.

        .. describe:: hash(x)

            Returns the stream's hash.

        .. describe:: str(x)

            Returns the stream's name.

    Attributes
    -----------
    platform: Optional[:class:`str`]
        Where the user is streaming from (ie. YouTube, Twitch).

        .. versionadded:: 1.3

    name: Optional[:class:`str`]
        The stream's name.
    details: Optional[:class:`str`]
        An alias for :attr:`name`
    game: Optional[:class:`str`]
        The game being streamed.

        .. versionadded:: 1.3

    url: :class:`str`
        The stream's URL.
    assets: :class:`dict`
        A dictionary comprising of similar keys than those in :attr:`Activity.assets`.
    """

    __slots__ = ('platform', 'name', 'game', 'url', 'details', 'assets')

    def __init__(self, *, name: Optional[str], url: str, **extra: Any) -> None:
        super().__init__(**extra)
        self.platform: Optional[str] = name
        self.name: Optional[str] = extra.pop('details', name)
        self.game: Optional[str] = extra.pop('state', None)
        self.url: str = url
        self.details: Optional[str] = extra.pop('details', self.name)  # compatibility
        self.assets: ActivityAssets = extra.pop('assets', {})

    @property
    def type(self) -> ActivityType:
        """:class:`ActivityType`: Returns the game's type. This is for compatibility with :class:`Activity`.

        It always returns :attr:`ActivityType.streaming`.
        """
        return ActivityType.streaming

    def __str__(self) -> str:
        return str(self.name)

    def __repr__(self) -> str:
        return f'<Streaming name={self.name!r}>'

    @property
    def twitch_name(self) -> Optional[str]:
        """Optional[:class:`str`]: If provided, the twitch name of the user streaming.

        This corresponds to the ``large_image`` key of the :attr:`Streaming.assets`
        dictionary if it starts with ``twitch:``. Typically set by the Discord client.
        """

        try:
            name = self.assets['large_image']
        except KeyError:
            return None
        else:
            return name[7:] if name[:7] == 'twitch:' else None

    def to_dict(self) -> ActivityPayload:
        ret: Dict[str, Any] = {
            'type': ActivityType.streaming.value,
            'name': str(self.name),
            'url': str(self.url),
            'assets': self.assets,
        }
        if self.details:
            ret['details'] = self.details
        return ret  # type: ignore

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Streaming) and other.name == self.name and other.url == self.url

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.name)


class Spotify:
    """Represents a Spotify listening activity from Discord. This is a special case of
    :class:`Activity` that makes it easier to work with the Spotify integration.

    .. container:: operations

        .. describe:: x == y

            Checks if two activities are equal.

        .. describe:: x != y

            Checks if two activities are not equal.

        .. describe:: hash(x)

            Returns the activity's hash.

        .. describe:: str(x)

            Returns the string 'Spotify'.
    """

    __slots__ = ('_state', '_details', '_timestamps', '_assets', '_party', '_sync_id', '_session_id', '_created_at')

    def __init__(self, **data: Any) -> None:
        self._state: str = data.pop('state', '')
        self._details: str = data.pop('details', '')
        self._timestamps: ActivityTimestamps = data.pop('timestamps', {})
        self._assets: ActivityAssets = data.pop('assets', {})
        self._party: ActivityParty = data.pop('party', {})
        self._sync_id: str = data.pop('sync_id', '')
        self._session_id: Optional[str] = data.pop('session_id')
        self._created_at: Optional[float] = data.pop('created_at', None)

    @property
    def type(self) -> ActivityType:
        """:class:`ActivityType`: Returns the activity's type. This is for compatibility with :class:`Activity`.

        It always returns :attr:`ActivityType.listening`.
        """
        return ActivityType.listening

    @property
    def created_at(self) -> Optional[datetime.datetime]:
        """Optional[:class:`datetime.datetime`]: When the user started listening in UTC.

        .. versionadded:: 1.3
        """
        if self._created_at is not None:
            return datetime.datetime.fromtimestamp(self._created_at / 1000, tz=datetime.timezone.utc)

    @property
    def colour(self) -> Colour:
        """:class:`Colour`: Returns the Spotify integration colour, as a :class:`Colour`.

        There is an alias for this named :attr:`color`"""
        return Colour(0x1DB954)

    @property
    def color(self) -> Colour:
        """:class:`Colour`: Returns the Spotify integration colour, as a :class:`Colour`.

        There is an alias for this named :attr:`colour`"""
        return self.colour

    def to_dict(self) -> ActivityPayload:
        return {
            'flags': 48,  # SYNC | PLAY
            'name': 'Spotify',
            'assets': self._assets,
            'party': self._party,
            'sync_id': self._sync_id,
            'session_id': self._session_id,
            'timestamps': self._timestamps,
            'details': self._details,
            'state': self._state,
        }  # type: ignore

    @property
    def name(self) -> str:
        """:class:`str`: The activity's name. This will always return "Spotify"."""
        return 'Spotify'

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Spotify)
            and other._session_id == self._session_id
            and other._sync_id == self._sync_id
            and other.start == self.start
        )

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self._session_id)

    def __str__(self) -> str:
        return 'Spotify'

    def __repr__(self) -> str:
        return f'<Spotify title={self.title!r} artist={self.artist!r} track_id={self.track_id!r}>'

    @property
    def title(self) -> str:
        """:class:`str`: The title of the song being played."""
        return self._details

    @property
    def artists(self) -> List[str]:
        """List[:class:`str`]: The artists of the song being played."""
        return self._state.split('; ')

    @property
    def artist(self) -> str:
        """:class:`str`: The artist of the song being played.

        This does not attempt to split the artist information into
        multiple artists. Useful if there's only a single artist.
        """
        return self._state

    @property
    def album(self) -> str:
        """:class:`str`: The album that the song being played belongs to."""
        return self._assets.get('large_text', '')

    @property
    def album_cover_url(self) -> str:
        """:class:`str`: The album cover image URL from Spotify's CDN."""
        large_image = self._assets.get('large_image', '')
        if large_image[:8] != 'spotify:':
            return ''
        album_image_id = large_image[8:]
        return 'https://i.scdn.co/image/' + album_image_id

    @property
    def track_id(self) -> str:
        """:class:`str`: The track ID used by Spotify to identify this song."""
        return self._sync_id

    @property
    def track_url(self) -> str:
        """:class:`str`: The track URL to listen on Spotify.

        .. versionadded:: 2.0
        """
        return f'https://open.spotify.com/track/{self.track_id}'

    @property
    def start(self) -> datetime.datetime:
        """:class:`datetime.datetime`: When the user started playing this song in UTC."""
        # the start key will be present here
        return datetime.datetime.fromtimestamp(self._timestamps['start'] / 1000, tz=datetime.timezone.utc)  # type: ignore

    @property
    def end(self) -> datetime.datetime:
        """:class:`datetime.datetime`: When the user will stop playing this song in UTC."""
        # the end key will be present here
        return datetime.datetime.fromtimestamp(self._timestamps['end'] / 1000, tz=datetime.timezone.utc)  # type: ignore

    @property
    def duration(self) -> datetime.timedelta:
        """:class:`datetime.timedelta`: The duration of the song being played."""
        return self.end - self.start

    @property
    def party_id(self) -> str:
        """:class:`str`: The party ID of the listening party."""
        return self._party.get('id', '')


class CustomActivity(BaseActivity):
    """Represents a custom activity from Discord.

    .. container:: operations

        .. describe:: x == y

            Checks if two activities are equal.

        .. describe:: x != y

            Checks if two activities are not equal.

        .. describe:: hash(x)

            Returns the activity's hash.

        .. describe:: str(x)

            Returns the custom status text.

    .. versionadded:: 1.3

    .. note::
        Technically, the name of custom activities is hardcoded to "Custom Status",
        and the state parameter has the actual custom text.
        This is confusing, so here, the name represents the actual custom text.
        However, the "correct" way of initialization still works.

    Attributes
    -----------
    name: Optional[:class:`str`]
        The custom activity's name.
    emoji: Optional[:class:`PartialEmoji`]
        The emoji to pass to the activity, if any.
    expires_at: Optional[:class:`datetime.datetime`]
        When the custom activity will expire. This is only available from :attr:`UserSettings.custom_activity`
    """

    __slots__ = ('name', 'emoji', 'expires_at')

    def __init__(
        self,
        name: Optional[str],
        *,
        emoji: Optional[PartialEmoji] = None,
        state: Optional[str] = None,
        expires_at: Optional[datetime.datetime] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if name == 'Custom Status':
            name = state
        self.name: Optional[str] = name
        self.expires_at = expires_at

        self.emoji: Optional[PartialEmoji]
        if isinstance(emoji, dict):
            self.emoji = PartialEmoji.from_dict(emoji)
        elif isinstance(emoji, str):
            self.emoji = PartialEmoji(name=emoji)
        elif isinstance(emoji, PartialEmoji) or emoji is None:
            self.emoji = emoji
        else:
            raise TypeError(f'Expected str, PartialEmoji, or None, received {type(emoji)!r} instead.')

    @classmethod
    def _from_legacy_settings(cls, *, data: Optional[dict], state: ConnectionState) -> Optional[Self]:
        if not data:
            return

        emoji = None
        if data.get('emoji_id'):
            emoji = state.get_emoji(int(data['emoji_id']))
            if not emoji:
                emoji = PartialEmoji(id=int(data['emoji_id']), name=data['emoji_name'])
                emoji._state = state
            else:
                emoji = emoji._to_partial()
        elif data.get('emoji_name'):
            emoji = PartialEmoji(name=data['emoji_name'])
            emoji._state = state

        return cls(name=data.get('text'), emoji=emoji, expires_at=parse_time(data.get('expires_at')))

    @classmethod
    def _from_settings(cls, *, data: Any, state: ConnectionState) -> Self:
        """
        message CustomStatus {
            string text = 1;
            fixed64 emoji_id = 2;
            string emoji_name = 3;
            fixed64 expires_at_ms = 4;
        }
        """
        emoji = None
        if data.emoji_id:
            emoji = state.get_emoji(data.emoji_id)
            if not emoji:
                emoji = PartialEmoji(id=data.emoji_id, name=data.emoji_name)
                emoji._state = state
            else:
                emoji = emoji._to_partial()
        elif data.emoji_name:
            emoji = PartialEmoji(name=data.emoji_name)
            emoji._state = state

        return cls(name=data.text, emoji=emoji, expires_at=parse_timestamp(data.expires_at_ms))

    @property
    def type(self) -> ActivityType:
        """:class:`ActivityType`: Returns the activity's type. This is for compatibility with :class:`Activity`.

        It always returns :attr:`ActivityType.custom`.
        """
        return ActivityType.custom

    def to_dict(self) -> ActivityPayload:
        payload = {
            'type': ActivityType.custom.value,
            'state': self.name,
            'name': 'Custom Status',  # Not a confusing API at all
        }
        if self.emoji:
            payload['emoji'] = self.emoji.to_dict()
        return payload  # type: ignore

    def to_legacy_settings_dict(self) -> Dict[str, Optional[Union[str, int]]]:
        payload: Dict[str, Optional[Union[str, int]]] = {}

        if self.name:
            payload['text'] = self.name
        if self.emoji:
            emoji = self.emoji
            payload['emoji_name'] = emoji.name
            if emoji.id:
                payload['emoji_id'] = emoji.id
        if self.expires_at is not None:
            payload['expires_at'] = self.expires_at.isoformat()
        return payload

    def to_settings_dict(self) -> Dict[str, Optional[Union[str, int]]]:
        payload: Dict[str, Optional[Union[str, int]]] = {}

        if self.name:
            payload['text'] = self.name
        if self.emoji:
            emoji = self.emoji
            payload['emoji_name'] = emoji.name
            if emoji.id:
                payload['emoji_id'] = emoji.id
        if self.expires_at is not None:
            payload['expires_at_ms'] = int(self.expires_at.timestamp() * 1000)
        return payload

    def __eq__(self, other: object) -> bool:
        return isinstance(other, CustomActivity) and other.name == self.name and other.emoji == self.emoji

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash((self.name, str(self.emoji)))

    def __str__(self) -> str:
        if self.emoji:
            if self.name:
                return f'{self.emoji} {self.name}'
            return str(self.emoji)
        else:
            return str(self.name)

    def __repr__(self) -> str:
        return f'<CustomActivity name={self.name!r} emoji={self.emoji!r}>'


class Session:
    """Represents a connected Discord gateway session.

    .. container:: operations

        .. describe:: x == y

            Checks if two sessions are equal.

        .. describe:: x != y

            Checks if two sessions are not equal.

        .. describe:: hash(x)

            Returns the session's hash.

    .. versionadded:: 2.0

    Attributes
    -----------
    session_id: :class:`str`
        The session ID.
    active: :class:`bool`
        Whether the session is active.
    os: :class:`OperatingSystem`
        The operating system the session is running on.
    client: :class:`ClientType`
        The client the session is running on.
    version: :class:`int`
        The version of the client the session is running on (used for differentiating between e.g. PS4/PS5).
    status: :class:`Status`
        The status of the session.
    activities: Tuple[Union[:class:`BaseActivity`, :class:`Spotify`]]
        The activities the session is currently doing.
    """

    __slots__ = (
        'session_id',
        'active',
        'os',
        'client',
        'version',
        'status',
        'activities',
        '_state',
    )

    def __init__(self, *, data: SessionPayload, state: ConnectionState):
        self._state = state
        client_info = data['client_info']

        self.session_id: str = data['session_id']
        self.os: OperatingSystem = OperatingSystem.from_string(client_info['os'])
        self.client: ClientType = try_enum(ClientType, client_info['client'])
        self.version: int = client_info.get('version', 0)
        self._update(data)

    def _update(self, data: SessionPayload):
        state = self._state

        # Only these should ever change
        self.active: bool = data.get('active', False)
        self.status: Status = try_enum(Status, data['status'])
        self.activities: Tuple[ActivityTypes, ...] = tuple(
            create_activity(activity, state) for activity in data['activities']
        )

    def __repr__(self) -> str:
        return f'<Session session_id={self.session_id!r} active={self.active!r} status={self.status!r} activities={self.activities!r}>'

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Session) and self.session_id == other.session_id

    def __ne__(self, other: object) -> bool:
        if isinstance(other, Session):
            return self.session_id != other.session_id
        return True

    def __hash__(self) -> int:
        return hash(self.session_id)

    @classmethod
    def _fake_all(cls, *, state: ConnectionState, data: SessionPayload) -> Self:
        self = cls.__new__(cls)
        self._state = state
        self.session_id = 'all'
        self.os = OperatingSystem.unknown
        self.client = ClientType.unknown
        self.version = 0
        self._update(data)
        return self

    def is_overall(self) -> bool:
        """:class:`bool`: Whether the session represents the overall presence across all platforms.

        .. note::

            If this is ``True``, then :attr:`session_id`, :attr:`os`, and :attr:`client` will not be real values.
        """
        return self.session_id == 'all'

    def is_headless(self) -> bool:
        """:class:`bool`: Whether the session is headless."""
        return self.session_id.startswith('h:')

    def is_current(self) -> bool:
        """:class:`bool`: Whether the session is the current session."""
        return self.session_id == self._state.session_id


ActivityTypes = Union[Activity, Game, CustomActivity, Streaming, Spotify]


@overload
def create_activity(data: ActivityPayload, state: ConnectionState) -> ActivityTypes:
    ...


@overload
def create_activity(data: None, state: ConnectionState) -> None:
    ...


@overload
def create_activity(data: Optional[ActivityPayload], state: ConnectionState) -> Optional[ActivityTypes]:
    ...


def create_activity(data: Optional[ActivityPayload], state: ConnectionState) -> Optional[ActivityTypes]:
    if not data:
        return None

    game_type = try_enum(ActivityType, data.get('type', -1))
    if game_type is ActivityType.playing:
        if 'application_id' in data or 'session_id' in data:
            return Activity(**data)
        return Game(**data)
    elif game_type is ActivityType.custom:
        try:
            name = data.pop('name')  # type: ignore
        except KeyError:
            ret = Activity(**data)
        else:
            # We removed the name key from data already
            ret = CustomActivity(name=name, **data)  # type: ignore
    elif game_type is ActivityType.streaming:
        if 'url' in data:
            # The url won't be None here
            return Streaming(**data)  # type: ignore
        return Activity(**data)
    elif game_type is ActivityType.listening and 'sync_id' in data and 'session_id' in data:
        return Spotify(**data)
    else:
        ret = Activity(**data)

    if isinstance(ret.emoji, PartialEmoji):
        ret.emoji._state = state
    return ret



"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import Literal, TypedDict
from typing_extensions import NotRequired


class BillingAddress(TypedDict):
    line_1: str
    line_2: NotRequired[str]
    name: str
    postal_code: NotRequired[str]
    city: str
    state: NotRequired[str]
    country: str
    email: NotRequired[str]


class BillingAddressToken(TypedDict):
    token: str


class PartialPaymentSource(TypedDict):
    id: str
    brand: NotRequired[str]
    country: NotRequired[str]
    last_4: NotRequired[str]
    type: Literal[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16]
    payment_gateway: Literal[1, 2, 3, 4, 5, 6]
    invalid: bool
    flags: int
    expires_month: NotRequired[int]
    expires_year: NotRequired[int]
    email: NotRequired[str]
    bank: NotRequired[str]
    username: NotRequired[str]
    screen_status: int  # TODO: Figure this out


class PaymentSource(PartialPaymentSource):
    billing_address: BillingAddress
    default: bool


class PremiumUsageValue(TypedDict):
    value: int


class PremiumUsage(TypedDict):
    nitro_sticker_sends: PremiumUsageValue
    total_animated_emojis: PremiumUsageValue
    total_global_emojis: PremiumUsageValue
    total_large_uploads: PremiumUsageValue
    total_hd_streams: PremiumUsageValue
    hd_hours_streamed: PremiumUsageValue



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import ClassVar, List, Literal, Optional, TYPE_CHECKING, Tuple, Union, overload

from .enums import try_enum, ComponentType, ButtonStyle, TextStyle, InteractionType
from .interactions import _wrapped_interaction
from .utils import _generate_nonce, get_slots, MISSING
from .partial_emoji import PartialEmoji, _EmojiTag

if TYPE_CHECKING:
    from typing_extensions import Self

    from .types.components import (
        Component as ComponentPayload,
        ButtonComponent as ButtonComponentPayload,
        SelectMenu as SelectMenuPayload,
        SelectOption as SelectOptionPayload,
        TextInput as TextInputPayload,
        ActionRowChildComponent as ActionRowChildComponentPayload,
    )
    from .emoji import Emoji
    from .interactions import Interaction
    from .message import Message

    ActionRowChildComponentType = Union['Button', 'SelectMenu', 'TextInput']


__all__ = (
    'Component',
    'ActionRow',
    'Button',
    'SelectMenu',
    'SelectOption',
    'TextInput',
)


class Component:
    """Represents a Discord Bot UI Kit Component.

    Currently, the only components supported by Discord are:

    - :class:`ActionRow`
    - :class:`Button`
    - :class:`SelectMenu`
    - :class:`TextInput`

    .. versionadded:: 2.0
    """

    __slots__ = ('message',)

    __repr_info__: ClassVar[Tuple[str, ...]]
    message: Message

    def __repr__(self) -> str:
        attrs = ' '.join(f'{key}={getattr(self, key)!r}' for key in self.__repr_info__)
        return f'<{self.__class__.__name__} {attrs}>'

    @property
    def type(self) -> ComponentType:
        """:class:`ComponentType`: The type of component."""
        raise NotImplementedError

    @classmethod
    def _raw_construct(cls, **kwargs) -> Self:
        self = cls.__new__(cls)
        for slot in get_slots(cls):
            try:
                value = kwargs[slot]
            except KeyError:
                pass
            else:
                setattr(self, slot, value)
        return self

    def to_dict(self) -> ComponentPayload:
        raise NotImplementedError


class ActionRow(Component):
    """Represents a Discord Bot UI Kit Action Row.

    This is a component that holds up to 5 children components in a row.

    This inherits from :class:`Component`.

    .. versionadded:: 2.0

    Attributes
    ------------
    children: List[Union[:class:`Button`, :class:`SelectMenu`, :class:`TextInput`]]
        The children components that this holds, if any.
    message: :class:`Message`
        The originating message.
    """

    __slots__ = ('children',)

    __repr_info__: ClassVar[Tuple[str, ...]] = __slots__

    def __init__(self, data: ComponentPayload, message: Message):
        self.message = message
        self.children: List[ActionRowChildComponentType] = []

        for component_data in data.get('components', []):
            component = _component_factory(component_data, message)

            if component is not None:
                self.children.append(component)

    @property
    def type(self) -> Literal[ComponentType.action_row]:
        """:class:`ComponentType`: The type of component."""
        return ComponentType.action_row


class Button(Component):
    """Represents a button from the Discord Bot UI Kit.

    This inherits from :class:`Component`.

    .. versionadded:: 2.0

    Attributes
    -----------
    style: :class:`.ButtonStyle`
        The style of the button.
    custom_id: Optional[:class:`str`]
        The ID of the button that gets received during an interaction.
        If this button is for a URL, it does not have a custom ID.
    url: Optional[:class:`str`]
        The URL this button sends you to.
    disabled: :class:`bool`
        Whether the button is disabled or not.
    label: Optional[:class:`str`]
        The label of the button, if any.
    emoji: Optional[:class:`PartialEmoji`]
        The emoji of the button, if available.
    message: :class:`Message`
        The originating message.
    """

    __slots__ = (
        'style',
        'custom_id',
        'url',
        'disabled',
        'label',
        'emoji',
    )

    __repr_info__: ClassVar[Tuple[str, ...]] = __slots__

    def __init__(self, data: ButtonComponentPayload, message: Message):
        self.message = message
        self.style: ButtonStyle = try_enum(ButtonStyle, data['style'])
        self.custom_id: Optional[str] = data.get('custom_id')
        self.url: Optional[str] = data.get('url')
        self.disabled: bool = data.get('disabled', False)
        self.label: Optional[str] = data.get('label')
        self.emoji: Optional[PartialEmoji]
        try:
            self.emoji = PartialEmoji.from_dict(data['emoji'])
        except KeyError:
            self.emoji = None

    @property
    def type(self) -> Literal[ComponentType.button]:
        """:class:`ComponentType`: The type of component."""
        return ComponentType.button

    def to_dict(self) -> dict:
        return {
            'component_type': self.type.value,
            'custom_id': self.custom_id,
        }

    async def click(self) -> Union[str, Interaction]:
        """|coro|

        Clicks the button.

        Raises
        -------
        InvalidData
            Didn't receive a response from Discord
            (doesn't mean the interaction failed).
        NotFound
            The originating message was not found.
        HTTPException
            Clicking the button failed.

        Returns
        --------
        Union[:class:`str`, :class:`Interaction`]
            The button's URL or the interaction that was created.
        """
        if self.url:
            return self.url

        message = self.message
        return await _wrapped_interaction(
            message._state,
            _generate_nonce(),
            InteractionType.component,
            None,
            message.channel,  # type: ignore # acc_channel is always correct here
            self.to_dict(),
            message=message,
        )


class SelectMenu(Component):
    """Represents a select menu from the Discord Bot UI Kit.

    A select menu is functionally the same as a dropdown, however
    on mobile it renders a bit differently.

    .. versionadded:: 2.0

    Attributes
    ------------
    custom_id: Optional[:class:`str`]
        The ID of the select menu that gets received during an interaction.
    placeholder: Optional[:class:`str`]
        The placeholder text that is shown if nothing is selected, if any.
    min_values: :class:`int`
        The minimum number of items that must be chosen for this select menu.
    max_values: :class:`int`
        The maximum number of items that must be chosen for this select menu.
    options: List[:class:`SelectOption`]
        A list of options that can be selected in this menu.
    disabled: :class:`bool`
        Whether the select is disabled or not.
    message: :class:`Message`
        The originating message, if any.
    """

    __slots__ = (
        'custom_id',
        'placeholder',
        'min_values',
        'max_values',
        'options',
        'disabled',
        'hash',
    )

    __repr_info__: ClassVar[Tuple[str, ...]] = __slots__

    def __init__(self, data: SelectMenuPayload, message: Message):
        self.message = message
        self.custom_id: str = data['custom_id']
        self.placeholder: Optional[str] = data.get('placeholder')
        self.min_values: int = data.get('min_values', 1)
        self.max_values: int = data.get('max_values', 1)
        self.options: List[SelectOption] = [SelectOption.from_dict(option) for option in data.get('options', [])]
        self.disabled: bool = data.get('disabled', False)
        self.hash: str = data.get('hash', '')

    @property
    def type(self) -> Literal[ComponentType.select]:
        """:class:`ComponentType`: The type of component."""
        return ComponentType.select

    def to_dict(self, options: Tuple[SelectOption]) -> dict:
        return {
            'component_type': self.type.value,
            'custom_id': self.custom_id,
            'values': [option.value for option in options],
        }

    async def choose(self, *options: SelectOption) -> Interaction:
        """|coro|

        Chooses the given options from the select menu.

        Raises
        -------
        InvalidData
            Didn't receive a response from Discord
            (doesn't mean the interaction failed).
        NotFound
            The originating message was not found.
        HTTPException
            Choosing the options failed.

        Returns
        --------
        :class:`Interaction`
            The interaction that was created.
        """
        message = self.message
        return await _wrapped_interaction(
            message._state,
            _generate_nonce(),
            InteractionType.component,
            None,
            message.channel,  # type: ignore # acc_channel is always correct here
            self.to_dict(options),
            message=message,
        )


class SelectOption:
    """Represents a select menu's option.

    .. versionadded:: 2.0

    Attributes
    -----------
    label: :class:`str`
        The label of the option. This is displayed to users.
        Can only be up to 100 characters.
    value: :class:`str`
        The value of the option. This is not displayed to users.
        If not provided when constructed then it defaults to the
        label. Can only be up to 100 characters.
    description: Optional[:class:`str`]
        An additional description of the option, if any.
        Can only be up to 100 characters.
    emoji: Optional[:class:`PartialEmoji`]
        The emoji of the option, if available.
    default: :class:`bool`
        Whether this option is selected by default.
    """

    __slots__ = (
        'label',
        'value',
        'description',
        'emoji',
        'default',
    )

    def __init__(
        self,
        *,
        label: str,
        value: str = MISSING,
        description: Optional[str] = None,
        emoji: Optional[Union[str, Emoji, PartialEmoji]] = None,
        default: bool = False,
    ) -> None:
        self.label: str = label
        self.value: str = label if value is MISSING else value
        self.description: Optional[str] = description

        if emoji is not None:
            if isinstance(emoji, str):
                emoji = PartialEmoji.from_str(emoji)
            elif isinstance(emoji, _EmojiTag):
                emoji = emoji._to_partial()
            else:
                raise TypeError(f'expected emoji to be str, Emoji, or PartialEmoji not {emoji.__class__}')

        self.emoji: Optional[PartialEmoji] = emoji
        self.default: bool = default

    def __repr__(self) -> str:
        return (
            f'<SelectOption label={self.label!r} value={self.value!r} description={self.description!r} '
            f'emoji={self.emoji!r} default={self.default!r}>'
        )

    def __str__(self) -> str:
        if self.emoji:
            base = f'{self.emoji} {self.label}'
        else:
            base = self.label

        if self.description:
            return f'{base}\n{self.description}'
        return base

    @classmethod
    def from_dict(cls, data: SelectOptionPayload) -> SelectOption:
        try:
            emoji = PartialEmoji.from_dict(data['emoji'])
        except KeyError:
            emoji = None

        return cls(
            label=data['label'],
            value=data['value'],
            description=data.get('description'),
            emoji=emoji,
            default=data.get('default', False),
        )


class TextInput(Component):
    """Represents a text input from the Discord Bot UI Kit.

    .. versionadded:: 2.0

    Attributes
    ------------
    custom_id: Optional[:class:`str`]
        The ID of the text input that gets received during an interaction.
    label: :class:`str`
        The label to display above the text input.
    style: :class:`TextStyle`
        The style of the text input.
    placeholder: Optional[:class:`str`]
        The placeholder text to display when the text input is empty.
    required: :class:`bool`
        Whether the text input is required.
    min_length: Optional[:class:`int`]
        The minimum length of the text input.
    max_length: Optional[:class:`int`]
        The maximum length of the text input.
    """

    __slots__ = (
        'style',
        'label',
        'custom_id',
        'placeholder',
        '_value',
        '_answer',
        'required',
        'min_length',
        'max_length',
    )

    __repr_info__: ClassVar[Tuple[str, ...]] = (
        'style',
        'label',
        'custom_id',
        'placeholder',
        'required',
        'min_length',
        'max_length',
        'default',
    )

    def __init__(self, data: TextInputPayload, *args) -> None:
        self.style: TextStyle = try_enum(TextStyle, data['style'])
        self.label: str = data['label']
        self.custom_id: str = data['custom_id']
        self.placeholder: Optional[str] = data.get('placeholder')
        self._value: Optional[str] = data.get('value')
        self.required: bool = data.get('required', True)
        self.min_length: Optional[int] = data.get('min_length')
        self.max_length: Optional[int] = data.get('max_length')

    @property
    def type(self) -> Literal[ComponentType.text_input]:
        """:class:`ComponentType`: The type of component."""
        return ComponentType.text_input

    @property
    def value(self) -> Optional[str]:
        """Optional[:class:`str`]: The current value of the text input. Defaults to :attr:`default`.

        This can be set to change the answer to the text input.
        """
        return getattr(self, '_answer', self._value)

    @value.setter
    def value(self, value: Optional[str]) -> None:
        length = len(value) if value is not None else 0
        if (self.required or value is not None) and (
            (self.min_length is not None and length < self.min_length)
            or (self.max_length is not None and length > self.max_length)
        ):
            raise ValueError(
                f'value cannot be shorter than {self.min_length or 0} or longer than {self.max_length or "infinity"}'
            )

        self._answer = value

    @property
    def default(self) -> Optional[str]:
        """Optional[:class:`str`]: The default value of the text input."""
        return self._value

    def answer(self, value: Optional[str], /) -> None:
        """A shorthand method to answer the text input.

        Parameters
        ----------
        value: Optional[:class:`str`]
            The value to set the answer to.

        Raises
        ------
        ValueError
            The answer is shorter than :attr:`min_length` or longer than :attr:`max_length`.
        """
        self.value = value

    def to_dict(self) -> dict:
        return {
            'type': self.type.value,
            'custom_id': self.custom_id,
            'value': self.value,
        }


@overload
def _component_factory(
    data: ActionRowChildComponentPayload, message: Message = ...
) -> Optional[ActionRowChildComponentType]:
    ...


@overload
def _component_factory(
    data: ComponentPayload, message: Message = ...
) -> Optional[Union[ActionRow, ActionRowChildComponentType]]:
    ...


def _component_factory(
    data: ComponentPayload, message: Message = MISSING
) -> Optional[Union[ActionRow, ActionRowChildComponentType]]:
    if data['type'] == 1:
        return ActionRow(data, message)
    elif data['type'] == 2:
        return Button(data, message)
    elif data['type'] == 3:
        return SelectMenu(data, message)
    elif data['type'] == 4:
        return TextInput(data, message)
